@@ -16,6 +16,43 @@ from .utils import (
 )
 
 
+@pytest.fixture
+def test_message_fixture(bouncer, pg):
+    yield bouncer, pg
+    pg.sql("ALTER USER test_error_message_user WITH LOGIN;")
+
+
+@pytest.mark.skipif("FREEBSD", reason="FreeBSD error reporting broken")
+@pytest.mark.skipif("WINDOWS", reason="Windows error reporting broken")
+def test_message(test_message_fixture):
+    bouncer, pg = test_message_fixture
+    test_user = "test_error_message_user"
+    connection_params = {"user": test_user, "dbname": "p0a"}
+    # Connect to database as User, creates existing pool
+    _ = bouncer.conn(**connection_params)
+
+    # Change user to nologin
+    pg.sql(f"ALTER USER {test_user} WITH NOLOGIN;")
+
+    # Kill process on postgres
+    terminate_string = f"""
+    SELECT pg_terminate_backend(pid)
+    FROM pg_stat_activity
+    WHERE
+      pid <> pg_backend_pid()
+      AND usename = '{test_user}'
+    """
+    pg.sql(terminate_string)
+
+    # login, check error message
+    # login again, check error message
+    for _ in range(2):
+        with pytest.raises(
+            psycopg.OperationalError, match=r"is not permitted to log in"
+        ):
+            bouncer.test(**connection_params)
+
+
 @pytest.mark.md5
 def test_auth_user(bouncer):
     bouncer.default_db = "authdb"
@@ -491,6 +528,43 @@ def test_auth_query_works_with_configured_users(bouncer):
                 )
 
 
+@pytest.mark.skipif("WINDOWS", reason="Windows does not have SIGHUP")
+def test_auth_query_logs_server_error(
+    bouncer,
+):
+    """
+    Check that when the auth_query response has an error, pgbouncer logs
+    the error message provided by postgres.
+    """
+
+    config = f"""
+        [databases]
+        postgres = auth_query='SELECT usename, passwd FROM not_pg_shadow where usename = $1'\
+            host={bouncer.pg.host} port={bouncer.pg.port}
+        [pgbouncer]
+        auth_query = SELECT 1
+        auth_user = pswcheck
+        stats_users = stats
+        listen_addr = {bouncer.host}
+        admin_users = pgbouncer
+        auth_type = md5
+        auth_file = {bouncer.auth_path}
+        listen_port = {bouncer.port}
+        logfile = {bouncer.log_path}
+        auth_dbname = postgres
+    """
+
+    with bouncer.log_contains('"not_pg_shadow" does not exist'):
+        with bouncer.run_with_config(config):
+            with pytest.raises(psycopg.OperationalError, match="bouncer config error"):
+                bouncer.sql(
+                    query="select version()",
+                    user="stats",
+                    password="stats",
+                    dbname="postgres",
+                )
+
+
 @pytest.mark.skipif("WINDOWS", reason="Windows does not have SIGHUP")
 @pytest.mark.md5
 def test_auth_dbname_works_fine(
@@ -604,7 +678,7 @@ async def test_change_server_password_reconnect(bouncer, pg):
             with pg.log_contains(
                 r"password authentication failed", times=1
             ), bouncer.log_contains(
-                r"closing because: password authentication failed for user", times=3
+                r"closing because: password authentication failed for user", times=4
             ):
                 result1 = bouncer.atest()
                 result2 = bouncer.atest()
@@ -626,6 +700,7 @@ async def test_change_server_password_reconnect(bouncer, pg):
                     psycopg.OperationalError, match="password authentication failed"
                 ):
                     await result3
+                time.sleep(3)
     finally:
         pg.sql("ALTER USER puser1 PASSWORD 'foo'")
 
@@ -649,7 +724,7 @@ async def test_change_server_password_server_lifetime(bouncer, pg):
         with pg.log_contains(
             r"password authentication failed", times=1
         ), bouncer.log_contains(
-            r"closing because: password authentication failed for user", times=3
+            r"closing because: password authentication failed for user", times=4
         ):
             result1 = bouncer.atest()
             result2 = bouncer.atest()
@@ -661,6 +736,7 @@ async def test_change_server_password_server_lifetime(bouncer, pg):
                 await result2
             with pytest.raises(psycopg.OperationalError):
                 await result3
+            time.sleep(3)
     finally:
         pg.sql("ALTER USER puser1 PASSWORD 'foo'")
 
