@@ -4,13 +4,13 @@ import time
 import psycopg
 import pytest
 from psycopg import pq, sql
+from psycopg.rows import dict_row
 
 from .utils import LIBPQ_SUPPORTS_PIPELINING, LINUX, PKT_BUF_SIZE, USE_SUDO
 
 
 def test_prepared_statement(bouncer):
     bouncer.admin(f"set pool_mode=transaction")
-    bouncer.admin(f"set max_prepared_statements=100")
     prepared_query = "SELECT 1"
     with bouncer.cur() as cur1:
         with bouncer.cur() as cur2:
@@ -32,7 +32,6 @@ def test_prepared_statement(bouncer):
 
 def test_prepared_statement_params(bouncer):
     bouncer.admin(f"set pool_mode=transaction")
-    bouncer.admin(f"set max_prepared_statements=100")
     prepared_query = "SELECT %s"
     with bouncer.cur() as cur1:
         with bouncer.cur() as cur2:
@@ -54,7 +53,6 @@ def test_prepared_statement_params(bouncer):
 
 def test_deallocate_all(bouncer):
     bouncer.admin(f"set pool_mode=transaction")
-    bouncer.admin(f"set max_prepared_statements=100")
     prepared_query = "SELECT 1"
     with bouncer.cur() as cur1:
         with bouncer.cur() as cur2:
@@ -84,7 +82,6 @@ def test_deallocate_all(bouncer):
 
 def test_discard_all(bouncer):
     bouncer.admin(f"set pool_mode=transaction")
-    bouncer.admin(f"set max_prepared_statements=100")
     prepared_query = "SELECT 1"
     with bouncer.cur() as cur1:
         with bouncer.cur() as cur2:
@@ -113,7 +110,6 @@ def test_discard_all(bouncer):
 
 
 def test_parse_larger_than_pkt_buf(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
     long_string = "1" * PKT_BUF_SIZE * 10
     prepared_query = "SELECT '" + long_string + "'"
     with bouncer.cur() as cur1:
@@ -122,7 +118,6 @@ def test_parse_larger_than_pkt_buf(bouncer):
 
 
 def test_bind_larger_than_pkt_buf(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
     long_string = "1" * PKT_BUF_SIZE * 10
     prepared_query = "SELECT %s::text"
     with bouncer.cur() as cur1:
@@ -136,7 +131,6 @@ def test_bind_larger_than_pkt_buf(bouncer):
 # buffer becomes larger than that it will be freed.
 # (see sbuf_process_pending in  sbuf.c)
 def test_parse_larger_than_pkt_buf_but_smaller_than_4x(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
     long_string = "1" * PKT_BUF_SIZE * 2
     prepared_query = "SELECT '" + long_string + "'"
     with bouncer.cur() as cur1:
@@ -145,7 +139,6 @@ def test_parse_larger_than_pkt_buf_but_smaller_than_4x(bouncer):
 
 
 def test_bind_larger_than_pkt_buf_but_smaller_than_4x(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
     long_string = "1" * PKT_BUF_SIZE * 2
     prepared_query = "SELECT %s::text"
     with bouncer.cur() as cur1:
@@ -159,7 +152,6 @@ def test_bind_larger_than_pkt_buf_but_smaller_than_4x(bouncer):
 # a bug, that if a varcache change was needed, then the callback would not be
 # called again correctly later.
 def test_parse_larger_than_pkt_buf_with_varcache_change(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
     long_string = "1" * PKT_BUF_SIZE * 10
     prepared_query = "SELECT '" + long_string + "'"
     with bouncer.cur(dbname="varcache_change") as cur1:
@@ -246,8 +238,6 @@ def test_evict_statement_cache_pipeline_failure(bouncer):
 
 @pytest.mark.skipif("not LIBPQ_SUPPORTS_PIPELINING")
 def test_prepared_statement_pipeline(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
-
     # Prepare query on the server connection and then disconnect again
     prepared_query = "SELECT 1"
     with bouncer.cur() as cur:
@@ -322,8 +312,6 @@ def test_describe_non_existent_prepared_statement(bouncer):
 # libpq before PG17 does not support sending Close messages
 @pytest.mark.skipif("psycopg.pq.version() < 170000")
 def test_close_prepared_statement(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
-
     with bouncer.conn() as conn:
         result = conn.pgconn.prepare(b"test", b"SELECT 1")
         assert result.status == pq.ExecStatus.COMMAND_OK
@@ -339,8 +327,6 @@ def test_close_prepared_statement(bouncer):
 
 
 def test_statement_name_longer_than_pkt_buf(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
-
     name = b"a" * PKT_BUF_SIZE * 4
 
     with bouncer.conn() as conn:
@@ -363,8 +349,6 @@ def test_statement_name_longer_than_pkt_buf(bouncer):
 
 @pytest.mark.skipif("not LIBPQ_SUPPORTS_PIPELINING")
 def test_prepared_statement_pipeline_error(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
-
     # Prepare query on the server connection and then disconnect again
     prepared_query = "SELECT 1"
     with bouncer.cur() as cur:
@@ -385,8 +369,6 @@ def test_prepared_statement_pipeline_error(bouncer):
 
 @pytest.mark.skipif("not LIBPQ_SUPPORTS_PIPELINING")
 def test_prepared_statement_pipeline_error_delayed_sync(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
-
     # Prepare query on the server connection and then disconnect again
     prepared_query = "SELECT 1"
     with bouncer.cur() as cur:
@@ -416,8 +398,6 @@ def test_prepared_statement_pipeline_error_delayed_sync(bouncer):
 
 
 def test_prepared_failed_prepare(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
-
     with bouncer.cur() as cur:
         with pytest.raises(psycopg.errors.UndefinedTable):
             cur.execute("SELECT * FROM doesnotexistyet", prepare=True)
@@ -428,8 +408,6 @@ def test_prepared_failed_prepare(bouncer):
 
 @pytest.mark.skipif("not LIBPQ_SUPPORTS_PIPELINING")
 def test_prepared_failed_prepare_pipeline(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
-
     with bouncer.conn() as conn:
         with conn.pipeline() as p, conn.cursor() as cur:
             cur.execute("SELECT 1", prepare=True)
@@ -448,8 +426,6 @@ def test_prepared_failed_prepare_pipeline(bouncer):
 
 
 def test_prepared_disallow_name_reuse(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
-
     with bouncer.conn() as conn:
         result = conn.pgconn.prepare(b"test", b"SELECT 1")
         assert result.status == pq.ExecStatus.COMMAND_OK
@@ -466,8 +442,6 @@ def test_prepared_disallow_name_reuse(bouncer):
 # more data.
 @pytest.mark.skipif("not LIBPQ_SUPPORTS_PIPELINING")
 def test_pipeline_with_half_pkt_buf_prepare(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
-
     long_string1 = "1" * (PKT_BUF_SIZE // 2)
     long_string2 = "2" * (PKT_BUF_SIZE // 2)
 
@@ -493,8 +467,6 @@ def test_pipeline_with_half_pkt_buf_prepare(bouncer):
 # data to do so.
 @pytest.mark.skipif("not LIBPQ_SUPPORTS_PIPELINING")
 def test_pipeline_flushes_on_full_pkt_buf(bouncer):
-    bouncer.admin(f"set max_prepared_statements=100")
-
     query = b"SELECT 1"
 
     # We want to construct a Parse packet that is exactly the size of pkt_buf,
@@ -598,3 +570,60 @@ def test_prepared_statement_pipeline_latency(bouncer, pg):
                 # The results should be correct too
                 for i in range(num_queries):
                     assert curs[i].fetchone()[0] == str(i)
+
+
+def test_prepared_statement_counters(bouncer):
+    bouncer.default_db = "p0"
+    bouncer.admin(f"set pool_mode=transaction")
+
+    # Explicitly disable prepared statement support
+    bouncer.admin(f"set max_prepared_statements=0")
+    with bouncer.cur() as cur:
+        with cur.connection.transaction():
+            cur.execute("SELECT 1")
+            cur.execute("SELECT 1")
+            cur.execute("SELECT 1")
+        with cur.connection.transaction():
+            cur.execute("SELECT 1")
+            cur.execute("SELECT 1")
+            cur.execute("SELECT 1")
+
+    stats = bouncer.admin("SHOW STATS", row_factory=dict_row)
+    p0_stats = next(s for s in stats if s["database"] == "p0")
+    # All the prepared statement counters should be 0, as they are not applicable
+    assert p0_stats["total_client_parse_count"] == 0
+    assert p0_stats["total_server_parse_count"] == 0
+    assert p0_stats["total_bind_count"] == 0
+    assert p0_stats["avg_client_parse_count"] == 0
+    assert p0_stats["avg_server_parse_count"] == 0
+    assert p0_stats["avg_bind_count"] == 0
+
+    # Explicitly enable prepared statement support
+    bouncer.admin(f"set max_prepared_statements=10")
+    prepared_query = "SELECT 1"
+    with bouncer.cur() as cur1:
+        with bouncer.cur() as cur2:
+            # prepare query on server 1 and client 1
+            cur1.execute(prepared_query, prepare=True)
+            # Run the prepared query twice on same server and client
+            cur1.execute(prepared_query)
+            cur1.execute(prepared_query)
+            with cur2.connection.transaction():
+                # Claim server 1 with client 2
+                cur2.execute("SELECT 2")
+                # Client 1 now runs the prepared query, and it's automatically
+                # prepared on server 2
+                cur1.execute(prepared_query)
+                # Client 2 now prepares the same query that was already
+                # prepared on server 1. And PgBouncer reuses that already
+                # prepared query for this different client.
+                cur2.execute(prepared_query, prepare=True)
+
+    stats = bouncer.admin("SHOW STATS", row_factory=dict_row)
+    p0_stats = next(s for s in stats if s["database"] == "p0")
+    # 2 prepare=True executions
+    assert p0_stats["total_client_parse_count"] == 2
+    # server 1 and server 2 have to prepare the query
+    assert p0_stats["total_server_parse_count"] == 2
+    # 2 executions with prepare=True + 3 re-use executions
+    assert p0_stats["total_bind_count"] == 5
