@@ -711,6 +711,10 @@ class Postgres(QueryRunner):
             # Make sure this is consistent across platforms
             pgconf.write("datestyle = 'iso, mdy'\n")
 
+            # Make PostgreSQL listen on both IPv4 and IPv6 (if supported)
+            if HAVE_IPV6_LOCALHOST:
+                pgconf.write("listen_addresses='127.0.0.1,::1'\n")
+
     def pgctl(self, command, **kwargs):
         run(f"pg_ctl -w --pgdata {self.pgdata} {command}", **kwargs)
 
